/* ============================================================================================== */
/*         Author       : Mark Streatfield                                                        */
/*         Contact      : a1490378@bournemouth.ac.uk                                              */
/* ============================================================================================== */
/*         File         : digit.c                                                                 */
/*         Version      : 1_04                                                                    */
/*         Date Created : 25-02-2004                                                              */
/* ============================================================================================== */
/*         About        : this code specifies a Linux based OpenGL program for the creation of 3D */
/*                        geometry from stereo image pairs.  This form of photogrammetry uses two */
/*                        images of a scene that are slightly offset to calculate a relative      */
/*                        depth for vertices in the image and so produce three dimensional data.  */
/*         Notes        : this code compilies under the ANSI C standard using the GNU C compiler  */
/*                        and the supplied makefile.  It has only been compiled and tested to run */
/*                        on machines running Red Hat Linux Enterprise Edition Version 3, but     */
/*                        no code has been identified as Linux specific and so should also        */
/*                        execute under a Windows operating system that has OpenGL support.       */
/*         Usage        : once compiled this program can be executed as follows                   */
/*                            ./digit <left_input_image> <right_input_image>                      */
/*                        where <left_input_image> and <right_input_image> are replaced with      */
/*                        valid paths to bitmap images                                            */
/*         References   : the code included in this source file has been created by myself.  The  */
/*                        code relating to the construction and maintainance of the linked lists  */
/*                        is based on that described in [Deitel, et al 1992] although has in      */
/*                        part been adapted for my own uses.  Other sources have been consulted   */
/*                        for research, most notably [Shreiner, et al 2003], however to my        */
/*                        knowledge no code has been directly used or copied                      */
/*         Known Bugs   : none, please post all bugs to a1490378@bournemouth.ac.uk                */
/* ============================================================================================== */

/* ============================================================================================== */
/*         Includes                                                                               */
/* ============================================================================================== */

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>

/* ============================================================================================== */
/*         Constant Definitions                                                                   */
/* ============================================================================================== */

#define IMAGE_WIDTH 360
#define IMAGE_HEIGHT 288
#define FALSE -1000
#define TRUE -2000

/* ============================================================================================== */
/*         Self-Defined Data Type Definitions                                                     */
/* ============================================================================================== */

/* in this data structure we store the start and end vertices of a line as taken from both the    */
/* left and right images, we ignore the y value of the right image to simplify the task           */
/* adapted from [Deitel, et al 1992]                                                              */

struct line_node{               /* a linked list node to hold data for lines/edges in the program */
	int start_left_x;
	int start_left_y;
	int start_right_x;
	float start_depth;
	int end_left_x;
	int end_left_y;
	int end_right_x;
	float end_depth;
	struct line_node *next_line_pointer;                          /* self-referential pointer */
};

/* in this data structure we stroe the coordinates of four vertices that are used to form the     */
/* corner of a cube, to reduce memory usage we do not store the right image coordinates           */
/* adapted from [Deitel, et al 1992]                                                              */

struct square_node{                   /* a linked list node to hold data for cubes in the program */
	int point_a_x;
	int point_a_y;
	float point_a_z;
	int point_b_x;
	int point_b_y;
	float point_b_z;
	int point_c_x;
	int point_c_y;
	float point_c_z;
	int point_d_x;
	int point_d_y;
	float point_d_z;
	struct square_node *next_square_pointer;                      /* self-referential pointer */
};

/* adapted from [Deitel, et al 1992]                                                              */
typedef struct line_node line_list_node;
typedef line_list_node *line_node_pointer;

typedef struct square_node square_list_node;
typedef square_list_node * square_node_pointer;

/* ============================================================================================== */
/*         Variable Declarations                                                                  */
/* ============================================================================================== */

line_node_pointer line_list_start = NULL;
square_node_pointer square_list_start = NULL;

/* these variables are used as boundaries for the left and right images so we can make sure that  */
/* the lines/edges/cubes created by the program fall within the image and are not erroneous       */

int left_image_bottom_position = FALSE;
int left_image_top_position = FALSE;
int left_image_left_position = FALSE;
int left_image_right_position = FALSE;
int right_image_bottom_position = FALSE;
int right_image_top_position = FALSE;
int right_image_left_position = FALSE;
int right_image_right_position = FALSE;

/* these variables are used to temporarily store the coordinates of a line/edge before being      */
/* stored in the linked list                                                                      */

int start_left_x_position = FALSE;
int start_left_y_position = FALSE;
int start_right_x_position = FALSE;
int end_left_x_position = FALSE;
int end_left_y_position = FALSE;
int end_right_x_position = FALSE;

/* these variables are used to temporarily store the coordinates of four vertices on a cube       */
/* before being stored in the linked list                                                         */

int left_image_point_a_x = FALSE;
int left_image_point_a_y = FALSE;
int left_image_point_b_x = FALSE;
int left_image_point_b_y = FALSE;
int left_image_point_c_x = FALSE;
int left_image_point_c_y = FALSE;
int left_image_point_d_x = FALSE;
int left_image_point_d_y = FALSE;
int right_image_point_a_x = FALSE;
int right_image_point_b_x = FALSE;
int right_image_point_c_x = FALSE;
int right_image_point_d_x = FALSE;

/* these are used as state variables to allow the program to determine the current state of       */
/* execution and ensure the appropiate function calls are made                                    */

int calibration = FALSE;
int add_line = FALSE;
int add_cube = FALSE;

/* variables defined here are used to store the current position of the virtual camera            */

float translate_x;
float translate_y;
float translate_z;
float rotate_x;
float rotate_y;
float zoom = 1.0;
float depth_factor = 1.0;
float scale = 1.0;

/* two pointers to a location in memory to store the bitmap image data                            */

GLubyte *left_image_array = NULL;
GLubyte *right_image_array = NULL;

/* OpenGL display list indexes                                                                    */

GLuint axis_display_list_index;
GLuint grid_display_list_index;

/* ============================================================================================== */
/* ============================================================================================== */
/* ============================================================================================== */
/* ============================================================================================== */

void save_bitmap(void)
{
GLubyte bgr_array[800][800][3];

FILE *output_file;

int i;
int j;
int image_size;
int image_width = 800;
int image_height = 800;
int image_x_resolution = 2519;
int image_y_resolution = 2519;
unsigned int image_file_size = 1920054;
unsigned int image_offset = 54;
unsigned int image_header_size = 40;
unsigned int image_compression = 0;
unsigned int image_image_size = 1920000;
unsigned int image_colours = 0;
unsigned int image_important_colours = 0;
unsigned short int image_type = 19778;
unsigned short int image_reserved1 = 0;
unsigned short int image_reserved2 = 0;
unsigned short int image_planes = 1;
unsigned short int image_bits_per_pixel = 24;
long bytes_per_line;
long actual_bytes_per_line;
short byte_alignment;

image_size = image_width * image_height * 3;

bytes_per_line = (image_width * image_bits_per_pixel) / 8;
actual_bytes_per_line = ((bytes_per_line + 3) / 4) * 4;
byte_alignment = actual_bytes_per_line - bytes_per_line;

output_file = fopen("./images/my_output.bmp", "wb");

if (!output_file)
	printf("Error : unable to open output file\n");

fwrite(&image_type, 2, 1, output_file);
fwrite(&image_file_size, 4, 1, output_file);
fwrite(&image_reserved1, 2, 1, output_file);
fwrite(&image_reserved2, 2, 1, output_file);
fwrite(&image_offset, 4, 1, output_file);
fwrite(&image_header_size, 4, 1, output_file);
fwrite(&image_width, 4, 1, output_file);
fwrite(&image_height, 4, 1, output_file);
fwrite(&image_planes, 2, 1, output_file);
fwrite(&image_bits_per_pixel, 2, 1, output_file);
fwrite(&image_compression, 4, 1, output_file);
fwrite(&image_image_size, 4, 1, output_file);
fwrite(&image_x_resolution, 4, 1, output_file);
fwrite(&image_y_resolution, 4, 1, output_file);
fwrite(&image_colours, 4, 1, output_file);
fwrite(&image_important_colours, 4, 1, output_file);

glReadPixels(0, 0, 800, 800, GL_BGR, GL_UNSIGNED_BYTE, bgr_array);

for (i = 0; i < image_height; i++)
{
	for (j = 0; j < image_width; j++)
		{
			fwrite(&bgr_array[i][j][0], 1, 1, output_file);
			fwrite(&bgr_array[i][j][1], 1, 1, output_file);
			fwrite(&bgr_array[i][j][2], 1, 1, output_file);
		}
		fseek(output_file, byte_alignment, SEEK_CUR);
}

fclose(output_file);
}

/* ============================================================================================== */
/* ============================================================================================== */
/* ============================================================================================== */
/* ============================================================================================== */


/* ============================================================================================== */
/*         Function Name : print_help                                                             */
/*         Function Args : none                                                                   */
/*         Function Task : prints help information to the shell prompt                            */
/* ============================================================================================== */

void print_help(void)
{
printf("      KEY                  ACTION             KEY                  ACTION\n");
printf("      -------------------------------------------------------------------------------------\n");
printf("      Esc                  terminate          c                    add cube\n");
printf("      Left Arrow           pan left           l                    add line/edge\n");
printf("      Right Arrow          pan right          o                    camera orientation\n");
printf("      Up Arrow             pan up             r                    recalibrate camera\n");
printf("      Down Arrow           pan down           d                    depth factor up\n");
printf("      Page Up              zoom in            f                    depth factor down\n");
printf("      Page Down            zoom out           .                    scale up\n");
printf("      a                    rotate x           ,                    scale down\n");
printf("      z                    rotate x           -                    zoom out\n");
printf("      v                    rotate y           =                    zoom in\n");
printf("      b                    rotate y           h                    display this help dialog\n");
printf("      -------------------------------------------------------------------------------------\n\n");
}

/* ============================================================================================== */
/*         Function Name : delete_line_list                                                       */
/*         Function Args : pointer to the start of the linked list containing edge data           */
/*         Function Task : deletes the entire linked list that contains the data for the lines/   */
/*                         edges in the scene                                                     */
/* ============================================================================================== */

void delete_line_list(line_node_pointer *list_start_pointer)
{
/* adapted from [Deitel, et al 1992]                                                              */
line_node_pointer previous_node = NULL;
line_node_pointer current_node = NULL;

if ((*list_start_pointer) != NULL)
	{
	previous_node = current_node;

	while(*list_start_pointer != NULL)
		{
		current_node = *list_start_pointer;

		if (current_node->next_line_pointer != NULL)
			*list_start_pointer = current_node->next_line_pointer;
		else
			*list_start_pointer = NULL;

		free(current_node);
		}
	*list_start_pointer = NULL;
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : delete_square_list                                                     */
/*         Function Args : pointer to the start of the linked list containing cube data           */
/*         Function Task : deletes the entire linked list that contains the data for the cubes    */
/*                         in the scene                                                           */
/* ============================================================================================== */

void delete_square_list(square_node_pointer *list_start_pointer)
{
/* adapted from [Deitel, et al 1992]                                                              */
square_node_pointer previous_node = NULL;
square_node_pointer current_node = NULL;

if ((*list_start_pointer) != NULL)
	{
	previous_node = current_node;

	while(*list_start_pointer != NULL)
		{
		current_node = *list_start_pointer;

		if (current_node->next_square_pointer != NULL)
			*list_start_pointer = current_node->next_square_pointer;
		else
			*list_start_pointer = NULL;

		free(current_node);
		}
	*list_start_pointer = NULL;
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : print_line_list                                                        */
/*         Function Args : the start of the linked list containing line data                      */
/*         Function Task : prints the data held in the linked list containing line/edge data      */
/* ============================================================================================== */

void print_line_list(line_node_pointer current_node)
{
/* adapted from [Deitel, et al 1992]                                                              */
if(current_node == NULL)
	printf("Error : no line list data to display\n");
else
	{
	printf("Digit : line list data ...\n");
	while (current_node != NULL)
		{
		printf("        ... NEW LINE START ...\n");
		printf("        ... start vertex left  image x = %d\n", current_node->start_left_x);
		printf("        ... start vertex left  image y = %d\n", current_node->start_left_y);
		printf("        ... start vertex right image x = %d\n", current_node->start_right_x);
		printf("        ... start vertex       depth   = %f\n", current_node->start_depth);
		printf("        ... end   vertex left  image x = %d\n", current_node->end_left_x);
		printf("        ... end   vertex left  image y = %d\n", current_node->end_left_y);
		printf("        ... end   vertex right image x = %d\n", current_node->end_right_x);
		printf("        ... end   vertex       depth   = %f\n", current_node->end_depth);
		printf("        ... CURRENT LINE END ...\n");

		current_node = current_node->next_line_pointer;
		}
	printf("Digit : ... end of line list data\n");
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : print_square_list                                                      */
/*         Function Args : the start of the linked list containing cube data                      */
/*         Function Task : prints the data held in the linked list containing cube data           */
/* ============================================================================================== */

void print_square_list(square_node_pointer current_node)
{
/* adapted from [Deitel, et al 1992]                                                              */
if (current_node == NULL)
	printf("Error : no cube list data to display\n");
else
	{
	printf("Digit : cube list data ...\n");
	while (current_node != NULL)
		{
		printf("        ... NEW CUBE START ...\n");
		printf("        ... vertex a coordinate x = %d\n", current_node->point_a_x);
		printf("        ... vertex a coordinate y = %d\n", current_node->point_a_y);
		printf("        ... vertex a coordinate z = %f\n", current_node->point_a_z);
		printf("        ... vertex b coordinate x = %d\n", current_node->point_b_x);
		printf("        ... vertex b coordinate y = %d\n", current_node->point_b_y);
		printf("        ... vertex b coordinate z = %f\n", current_node->point_b_z);
		printf("        ... vertex c coordinate x = %d\n", current_node->point_c_x);
		printf("        ... vertex c coordinate y = %d\n", current_node->point_c_y);
		printf("        ... vertex c coordinate z = %f\n", current_node->point_c_z);
		printf("        ... vertex d coordinate x = %d\n", current_node->point_d_x);
		printf("        ... vertex d coordinate y = %d\n", current_node->point_d_y);
		printf("        ... vertex d coordinate z = %f\n", current_node->point_d_z);
		printf("        ... CURRENT CUBE END ...\n");

		current_node = current_node->next_square_pointer;
		}
	printf("Digit : ... end of cube list data\n");
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : exit_func                                                              */
/*         Function Args : none                                                                   */
/*         Function Task : executed whenever an "exit()" call is made to ensure clean             */
/*                         terminateion of program                                                */
/* ============================================================================================== */

void exit_func(void)
{
if (left_image_array != NULL); /* free the memory associated with the left image                  */
	free(left_image_array);

if (right_image_array != NULL); /* free the memory associated with the right image                */
	free(right_image_array);

#ifdef DEBUG
	print_line_list(line_list_start);
	print_square_list(square_list_start);
#endif

delete_line_list(&line_list_start); /* free the memory associated with the edge/line linked list  */
delete_square_list(&square_list_start); /* free the memory associated with the cube linked list   */

#ifdef DEBUG
	print_line_list(line_list_start);
	print_square_list(square_list_start);
#endif

printf("Digit : terminating, good bye\n\n");
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : load_images                                                            */
/*         Function Args : pointer to a character string containing the file names for the left   */
/*                         and right image to be loaded                                           */
/*         Function Task : reads the byte data of the images from disk to memory to be displayed  */
/*                         in OpenGL                                                              */
/* ============================================================================================== */

int load_images(char *left_filename, char *right_filename)
{
FILE *left_input_file = NULL;
FILE *right_input_file = NULL;

long left_image_bytes_per_line;
long left_image_actual_bytes_per_line;
short left_image_byte_alignment;
long right_image_bytes_per_line;
long right_image_actual_bytes_per_line;
short right_image_byte_alignment;

unsigned short int left_image_type;
unsigned short int right_image_type;
unsigned short int left_image_planes;
unsigned short int right_image_planes;
unsigned short int left_image_bits_per_pixel;
unsigned short int right_image_bits_per_pixel;
unsigned int left_image_compression;
unsigned int right_image_compression;
int left_image_width;
int right_image_width;
int left_image_height;
int right_image_height;
int left_image_size;
int right_image_size;
int i;

/* because the program, at this stage, is very specific as to the images it will accept - only    */
/* bitmaps that are 360x288 pixels - a lot of checking is required to ensure that the image files */
/* conform to these criteria.  As a result more code is used than strictly necessary but provides */
/* a "bolt and braces" approach to avoid errors                                                   */

left_input_file = fopen(left_filename, "rb"); /* open the left image file for reading             */
right_input_file = fopen(right_filename, "rb"); /* open the right image file for reading          */

if (!left_input_file) /* check success and terminate on failure                                   */
	{
	printf("Error : unable to open left image source\n");
	return 0;
	}
else if (!right_input_file)
	{
	fclose(left_input_file);
	printf("Error : unable to open right image source\n");
	return 0;
	}

/* read in the file identifier from the image file and confirm it is a bitmap image               */
fread(&left_image_type, 2, 1, left_input_file);
fread(&right_image_type, 2, 1, right_input_file);

if (left_image_type != 19778)
	{
	printf("Error : incorrect source type for left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_type != 19778)
	{
	printf("Error : incorrect source type for right image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* more through the bitmap header to were the image width and height is stored                    */
fseek(left_input_file, 16, SEEK_CUR);
fseek(right_input_file, 16, SEEK_CUR);

/* read in the width and height and confirm the images are of the correct size                    */
fread(&left_image_width, 4, 1, left_input_file);
fread(&left_image_height, 4, 1, left_input_file);
fread(&right_image_width, 4, 1, right_input_file);
fread(&right_image_height, 4, 1, right_input_file);

if (left_image_width != IMAGE_WIDTH || left_image_height != IMAGE_HEIGHT)
	{
	printf("Error : incorrect image size in left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_width != IMAGE_WIDTH || right_image_height != IMAGE_HEIGHT)
	{
	printf("Error : incorrect image size in right image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* calculate the size of the images based on their height, width and rgb values                   */
left_image_size = left_image_width * left_image_height * 3;
right_image_size = right_image_width * right_image_height * 3;

/* read in the number of planes in the bitmap image and ensure they equal 1                       */
fread(&left_image_planes, 2, 1, left_input_file);
fread(&right_image_planes, 2, 1, right_input_file);

if (left_image_planes != 1)
	{
	printf("Error : incorrect image format in left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_planes != 1)
	{
	printf("Error : incorrect image format in right image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* ensure the images have 24 bits per pixel                                                       */
fread(&left_image_bits_per_pixel, 2, 1, left_input_file);
fread(&right_image_bits_per_pixel, 2, 1, right_input_file);

if (left_image_bits_per_pixel != 24)
	{
	printf("Error : incorrect image formate in left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_bits_per_pixel != 24)
	{
	printf("Error : incorrect image format in right image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* confirm that there is no compression on the images                                             */
fread(&left_image_compression, 4, 1, left_input_file);
fread(&right_image_compression, 4, 1, right_input_file);

if (left_image_compression != 0)
	{
	printf("Error : incorrect image formate in left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_compression != 0)
	{
	printf("Error : incorrect image format in right image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* move to the file pointer to point to a total of 54 bytes within the file where the rgb byte    */
/* data is stored                                                                                 */
fseek(left_input_file, 20, SEEK_CUR);
fseek(right_input_file, 20, SEEK_CUR);

/* calculate the number of junk bytes, or byte alignment, stored after the rgb data on each scan  */
/* line of the image, if this is not done the image will be corrupt                               */
left_image_bytes_per_line = (left_image_width * left_image_bits_per_pixel) / 8;
right_image_bytes_per_line = (right_image_width * right_image_bits_per_pixel) / 8;

left_image_actual_bytes_per_line = ((left_image_bytes_per_line + 3) / 4) * 4;
right_image_actual_bytes_per_line = ((right_image_bytes_per_line + 3) / 4) * 4;

left_image_byte_alignment = left_image_actual_bytes_per_line - left_image_bytes_per_line;
right_image_byte_alignment = right_image_actual_bytes_per_line - right_image_bytes_per_line;

/* allocate memory for the left and right images and test success                                 */
left_image_array = (GLubyte *) malloc(left_image_width * left_image_height * 3);
right_image_array = (GLubyte *) malloc(right_image_width * right_image_height * 3);

if (left_image_array == NULL)
	{
	printf("Error : memory allocation failure in left image\n");
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}
else if (right_image_array == NULL)
	{
	printf("Error : memory allocation failure in right image\n");
	free(left_image_array);
	fclose(left_input_file);
	fclose(right_input_file);
	return 0;
	}

/* read in the byte data from the left image file                                                 */
for (i = 0; i < left_image_height; i++)
	{
	fread((left_image_array + i * left_image_width * 3), left_image_width * 3, 1, left_input_file);
	fseek(left_input_file, left_image_byte_alignment, SEEK_CUR);
	}

/* read in the byte data from the right image file                                                */
for (i = 0; i < right_image_height; i++)
	{
	fread((right_image_array + i * right_image_width * 3), right_image_width * 3, 1, right_input_file);
	fseek(right_input_file, right_image_byte_alignment, SEEK_CUR);
	}

fclose(left_input_file); /* close both input files                                                */
fclose(right_input_file);

return 1;
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : axis_display_list                                                      */
/*         Function Args : none                                                                   */
/*         Function Task : creates a display list for a virtual axis that is displayed in OpenGL  */
/* ============================================================================================== */

void axis_display_list(void)
{
axis_display_list_index = glGenLists(1);

glNewList(axis_display_list_index, GL_COMPILE);
	glLineWidth(3.0);

	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(5.0, 0.0, 0.0);
	glEnd();

	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 5.0, 0.0);
	glEnd();

	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 0.0, 5.0);
	glEnd();

	glLineWidth(1.0);
glEndList();
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : grid_display_list                                                      */
/*         Function Args : none                                                                   */
/*         Function Task : creates a display list for a virtual grid that is displayed in OpenGL  */
/* ============================================================================================== */

void grid_display_list(void)
{
int i;

grid_display_list_index = glGenLists(1);

glNewList(grid_display_list_index, GL_COMPILE);
	glColor3f(0.715, 0.715, 0.715);

	for (i = 0; i <= 50; i++)
		{
		glBegin(GL_LINES);
			glVertex3f(-50.0, 0.0, (i * 2)-50);
			glVertex3f(50.0, 0.0, (i * 2)-50);
		glEnd();
		}

	for (i = 0; i <= 50; i++)
		{
		glBegin(GL_LINES);
			glVertex3f((i * 2)-50, 0.0, -50.0);
			glVertex3f((i * 2)-50, 0.0, 50.0);
		glEnd();
		}
glEndList();
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : insert_line                                                            */
/*         Function Args : none                                                                   */
/*         Function Task : adds a new node to the line/edge linked list                           */
/* ============================================================================================== */

void insert_line(void)
{
/* adapted from [Deitel, et al 1992]                                                              */
line_node_pointer new_node = NULL;
line_node_pointer previous_node = NULL;
line_node_pointer current_node = NULL;

new_node = malloc(sizeof(line_list_node));

if (new_node == NULL)
	{
	printf("Error : memory allocation failure in new line\n");
	exit(0);
	}

new_node->start_left_x = start_left_x_position - left_image_left_position;
new_node->start_left_y = left_image_bottom_position - start_left_y_position;
new_node->start_right_x = start_right_x_position - right_image_left_position;

new_node->start_depth = new_node->start_left_x - new_node->start_right_x;

new_node->end_left_x = end_left_x_position - left_image_left_position;
new_node->end_left_y = left_image_bottom_position - end_left_y_position;
new_node->end_right_x = end_right_x_position - right_image_left_position;

new_node->end_depth = new_node->end_left_x - new_node->end_right_x;

new_node->next_line_pointer = NULL;
previous_node = NULL;
current_node = line_list_start;

while(current_node != NULL)
	{
	previous_node = current_node;
	current_node = current_node->next_line_pointer;
	}

if (previous_node == NULL)
	{
	new_node->next_line_pointer = line_list_start;
	line_list_start = new_node;
	}
else
	{
	previous_node->next_line_pointer = new_node;
	new_node->next_line_pointer = current_node;
	}

printf("Digit : inserting new line based on following data ...\n");
printf("        ... start vertex left  image x = %d\n", new_node->start_left_x);
printf("        ... start vertex left  image y = %d\n", new_node->start_left_y);
printf("        ... start vertex right image x = %d\n", new_node->start_right_x);
printf("        ... start vertex       depth   = %f\n", new_node->start_depth);
printf("        ... end   vertex left  image x = %d\n", new_node->end_left_x);
printf("        ... end   vertex left  image y = %d\n", new_node->end_left_y);
printf("        ... end   vertex right image x = %d\n", new_node->end_right_x);
printf("        ... end   vertex       depth   = %f\n", new_node->end_depth);
printf("Digit : ... insert complete\n");
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : insert_square                                                          */
/*         Function Args : none                                                                   */
/*         Function Task : adds a new node to the cube linked list                                */
/* ============================================================================================== */

void insert_square(void)
{
/* adapted from [Deitel, et al 1992]                                                              */
square_node_pointer new_node = NULL;
square_node_pointer previous_node = NULL;
square_node_pointer current_node = NULL;

new_node = malloc(sizeof(square_list_node));

if (new_node == NULL)
	{
	printf("Error : memory allocation failure in new square\n");
	exit(0);
	}

new_node->point_a_x = left_image_point_a_x - left_image_left_position;
new_node->point_a_y = left_image_bottom_position - left_image_point_a_y;
new_node->point_a_z = new_node->point_a_x - (right_image_point_a_x - right_image_right_position);

new_node->point_b_x = left_image_point_b_x - left_image_left_position;
new_node->point_b_y = left_image_bottom_position - left_image_point_b_y;
new_node->point_b_z = new_node->point_b_x - (right_image_point_b_x - right_image_right_position);

new_node->point_c_x = left_image_point_c_x - left_image_left_position;
new_node->point_c_y = left_image_bottom_position - left_image_point_c_y;
new_node->point_c_z = new_node->point_c_x - (right_image_point_c_x - right_image_right_position);

new_node->point_d_x = left_image_point_d_x - left_image_left_position;
new_node->point_d_y = left_image_bottom_position - left_image_point_d_y;
new_node->point_d_z = new_node->point_d_x - (right_image_point_d_x - right_image_right_position);

new_node->next_square_pointer = NULL;
previous_node = NULL;
current_node = square_list_start;

while(current_node != NULL)
	{
	previous_node = current_node;
	current_node = current_node->next_square_pointer;
	}

if (previous_node == NULL)
	{
	new_node->next_square_pointer = square_list_start;
	square_list_start = new_node;
	}
else
	{
	previous_node->next_square_pointer = new_node;
	new_node->next_square_pointer = current_node;
	}

printf("Digit : inserting new cube based on following data ...\n");
printf("        ... vertex a coordinate x = %d\n", new_node->point_a_x);
printf("        ... vertex a coordinate y = %d\n", new_node->point_a_y);
printf("        ... vertex a coordinate z = %f\n", new_node->point_a_z);
printf("        ... vertex b coordinate x = %d\n", new_node->point_b_x);
printf("        ... vertex b coordinate y = %d\n", new_node->point_b_y);
printf("        ... vertex b coordinate z = %f\n", new_node->point_b_z);
printf("        ... vertex c coordinate x = %d\n", new_node->point_c_x);
printf("        ... vertex c coordinate y = %d\n", new_node->point_c_y);
printf("        ... vertex c coordinate z = %f\n", new_node->point_c_z);
printf("        ... vertex d coordinate x = %d\n", new_node->point_d_x);
printf("        ... vertex d coordinate y = %d\n", new_node->point_d_y);
printf("        ... vertex d coordinate z = %f\n", new_node->point_d_z);
printf("Digit : ... insert complete\n");
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : display_object                                                         */
/*         Function Args : the start of the two linked lists used in the code                     */
/*         Function Task : completely traverse the two linked lists, interpreting the data and    */
/*                         displaying it in OpenGL                                                */
/* ============================================================================================== */

void display_object(line_node_pointer current_node, square_node_pointer current_square_node)
{
int horizontal_adjustment;
int vertical_adjustment;

int cube_depth;
int cube_width;
int cube_breadth;

horizontal_adjustment = (left_image_right_position - left_image_left_position) * 0.5;
vertical_adjustment = (left_image_bottom_position - left_image_top_position) * 0.5;

glColor3f(0.0, 0.965, 1.0);

glLineWidth(2.0);

/* traverse through the complete linked list that contains the line/edge data and display the     */
/* lines on screen.  The horizontal and vertical adjustment values are used to move the center of */
/* the scene to the origin, instead of having the bottom left corner centered at the origin       */
if (current_node != NULL)
	{
	while(current_node != NULL)
		{
		glBegin(GL_LINES);
			glVertex3f((current_node->start_left_x - horizontal_adjustment) * scale,
					(current_node->start_left_y - vertical_adjustment) * scale,
					current_node->start_depth * depth_factor * scale);
			glVertex3f((current_node->end_left_x - horizontal_adjustment) * scale,
					(current_node->end_left_y - vertical_adjustment) * scale,
					current_node->end_depth * depth_factor * scale);
		glEnd();

		current_node = current_node->next_line_pointer;
		}
	}

/* traverse through the complete linked list that contains the cube data and display the cube on  */
/* screen.  The lines connecting the four vertices that were picked by the user are displayed in  */
/* red while the green lines represent those that the code reconstructs to show the complete cube */
/* This code is still under construction and as such is not complete.                             */
if (current_square_node != NULL)
	{
	while(current_square_node != NULL)
		{
		glColor3f(0.580, 1.0, 0.478);

		cube_depth = current_square_node->point_b_y - current_square_node->point_d_y;
		cube_width = current_square_node->point_a_x - current_square_node->point_b_x ;
		cube_breadth = current_square_node->point_c_z - current_square_node->point_b_z ;

		glBegin(GL_LINES); /* draw line AB                                                */
			glVertex3f((current_square_node->point_a_x - horizontal_adjustment) * scale,
					(current_square_node->point_a_y - vertical_adjustment) * scale,
					current_square_node->point_a_z * depth_factor * scale);
			glVertex3f((current_square_node->point_b_x - horizontal_adjustment) * scale,
					(current_square_node->point_b_y - vertical_adjustment) * scale,
					current_square_node->point_b_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line BC                                                */
			glVertex3f((current_square_node->point_b_x - horizontal_adjustment) * scale,
					(current_square_node->point_b_y - vertical_adjustment) * scale,
					current_square_node->point_b_z * depth_factor * scale);
			glVertex3f((current_square_node->point_c_x - horizontal_adjustment) * scale,
					(current_square_node->point_c_y - vertical_adjustment) * scale,
					current_square_node->point_c_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line BD                                                */
			glVertex3f((current_square_node->point_b_x - horizontal_adjustment) * scale,
					(current_square_node->point_b_y - vertical_adjustment) * scale,
					current_square_node->point_b_z * depth_factor * scale);
			glVertex3f((current_square_node->point_d_x - horizontal_adjustment) * scale,
					(current_square_node->point_d_y - vertical_adjustment) * scale,
					current_square_node->point_d_z * depth_factor * scale);
		glEnd();

		glColor3f(1.0, 0.318, 0.318);

		/* code beyond this point has not been completed fully, it has been included      */
		/* because when executed it is possible to see the potential of this method       */
		glBegin(GL_LINES); /* draw line DE                                                */
			glVertex3f((current_square_node->point_d_x - horizontal_adjustment) * scale,
					(current_square_node->point_d_y - vertical_adjustment) * scale,
					current_square_node->point_d_z * depth_factor * scale);
			glVertex3f((current_square_node->point_a_x - horizontal_adjustment) * scale,
					((current_square_node->point_a_y - cube_depth) - vertical_adjustment) * scale,
					current_square_node->point_a_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line AE                                                */
			glVertex3f((current_square_node->point_a_x - horizontal_adjustment) * scale,
					(current_square_node->point_a_y - vertical_adjustment) * scale,
					current_square_node->point_a_z * depth_factor * scale);
			glVertex3f((current_square_node->point_a_x - horizontal_adjustment) * scale,
					((current_square_node->point_a_y - cube_depth) - vertical_adjustment) * scale,
					current_square_node->point_a_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line DG                                                */
			glVertex3f((current_square_node->point_d_x - horizontal_adjustment) * scale,
					(current_square_node->point_d_y - vertical_adjustment) * scale,
					current_square_node->point_d_z * depth_factor * scale);
			glVertex3f((current_square_node->point_c_x - horizontal_adjustment) * scale,
					((current_square_node->point_c_y -cube_depth) - vertical_adjustment) * scale,
					current_square_node->point_c_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line CG                                                */
			glVertex3f((current_square_node->point_c_x - horizontal_adjustment) * scale,
					(current_square_node->point_c_y - vertical_adjustment) * scale,
					current_square_node->point_c_z * depth_factor * scale);
			glVertex3f((current_square_node->point_c_x - horizontal_adjustment) * scale,
					((current_square_node->point_c_y -cube_depth) - vertical_adjustment) * scale,
					current_square_node->point_c_z * depth_factor * scale);
		glEnd();

		glBegin(GL_LINES); /* draw line AF                                                */
		glEnd();

		glBegin(GL_LINES); /* draw line CF                                                */
		glEnd();

		glBegin(GL_LINES); /* draw line FH                                                */
		glEnd();

		glBegin(GL_LINES); /* draw line EH                                                */
		glEnd();

		glBegin(GL_LINES); /* draw line GH                                                */
		glEnd();

		current_square_node = current_square_node->next_square_pointer;
		}
	}

glLineWidth(1.0);
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : display_func                                                           */
/*         Function Args : none                                                                   */
/*         Function Task : this is the registered display function for OpenGL and writes to the   */
/*                         display buffer                                                         */
/* ============================================================================================== */

void display_func(void)
{
glEnable(GL_DEPTH_TEST);

glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

glMatrixMode(GL_MODELVIEW);
glLoadIdentity();
gluLookAt(0.0, 0.0, 12.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

glPushMatrix();

/* the following calls to "glRotate" "glTranslate" and "glScale" move the scene relative to the   */
/* virtual camera, giving the appearance that the camera has or is moving                         */
glRotatef(rotate_x, 1.0, 0.0, 0.0 ) ;
glRotatef(rotate_y, 0.0, 1.0, 0.0 ) ;

glTranslatef(translate_x, translate_y, translate_z);

glScalef(zoom, zoom, zoom);

glCallList(axis_display_list_index); /* draw the axis using the display list                      */

glCallList(grid_display_list_index); /* draw the grid using the display list                      */

display_object(line_list_start, square_list_start);

glPopMatrix();

glDisable(GL_DEPTH_TEST);

glRasterPos2f(-5.5,-5.5); /* display the left image on screen                                     */
glDrawPixels(IMAGE_WIDTH, IMAGE_HEIGHT, GL_BGR, GL_UNSIGNED_BYTE, left_image_array);
glRasterPos2f(0.25,-5.5); /* display the right image on screen                                    */
glDrawPixels(IMAGE_WIDTH, IMAGE_HEIGHT, GL_BGR, GL_UNSIGNED_BYTE, right_image_array);

/* the execution of the program is determind by the current state, and the user is prompted       */
/* accordingly.  If the state is not "add line", "add cube" or "calibrate" then this section is   */
/* bypassed, otherwise the execution steps through this "if" statement on each redisplay          */

/* if the images are being calibrated ...                                                         */
if (calibration == FALSE)
	{
	glutSwapBuffers();

	/* determine the current stage in the calibration process and prompt the user for the     */
	/* point to be calibrated                                                                 */
	if (left_image_bottom_position == FALSE && left_image_top_position == FALSE &&
			left_image_left_position == FALSE && left_image_right_position == FALSE &&
			right_image_bottom_position == FALSE && right_image_top_position == FALSE &&
			right_image_left_position == FALSE && right_image_right_position == FALSE)
		{
		printf("Digit : calibrate camera ...\n");
		printf("Digit : click on bottom left pixel of left image ...\n");
		}
	else if	(left_image_bottom_position != FALSE && left_image_top_position == FALSE &&
			left_image_left_position != FALSE && left_image_right_position == FALSE &&
			right_image_bottom_position == FALSE && right_image_top_position == FALSE &&
			right_image_left_position == FALSE && right_image_right_position == FALSE)
		printf("Digit : click on top right pixel of left image ...\n");
	else if	(left_image_bottom_position != FALSE && left_image_top_position != FALSE &&
			left_image_left_position != FALSE && left_image_right_position != FALSE &&
			right_image_bottom_position == FALSE && right_image_top_position == FALSE &&
			right_image_left_position == FALSE && right_image_right_position == FALSE)
		printf("Digit : click on bottom left pixel of right image ...\n");
	else if	(left_image_bottom_position != FALSE && left_image_top_position != FALSE &&
			left_image_left_position != FALSE && left_image_right_position != FALSE &&
			right_image_bottom_position != FALSE && right_image_top_position == FALSE &&
			right_image_left_position != FALSE && right_image_right_position == FALSE)
		printf("Digit : click on top right pixel of right image ...\n");
	}
/* if a new line/edge is being added ...                                                          */
else if (add_line == TRUE)
	{
	glutSwapBuffers();

	/* determine the current stage in the addition of a new line/edge and prompt the user for */
	/* the next vertex in either the left of right image                                      */
	if (start_left_x_position == FALSE && start_left_y_position == FALSE &&
			start_right_x_position == FALSE && end_left_x_position == FALSE &&
			end_left_y_position == FALSE && end_right_x_position == FALSE)
		{
		printf("Digit : add line ...\n");
		printf("Digit : click on start point of first vertex in left image ...\n");
		}
	else if (start_left_x_position != FALSE && start_left_y_position != FALSE &&
			start_right_x_position == FALSE && end_left_x_position == FALSE &&
			end_left_y_position == FALSE && end_right_x_position == FALSE)
		printf("Digit : click on end point of first vertex in right image ...\n");
	else if (start_left_x_position != FALSE && start_left_y_position != FALSE &&
			start_right_x_position != FALSE && end_left_x_position == FALSE &&
			end_left_y_position == FALSE && end_right_x_position == FALSE)
		printf("Digit : click on start point of second vertex in left image ...\n");
	else if (start_left_x_position != FALSE && start_left_y_position != FALSE &&
			start_right_x_position != FALSE && end_left_x_position != FALSE &&
			end_left_y_position != FALSE && end_right_x_position == FALSE)
		printf("Digit : click on end point of second vertex in right image ...\n");
	}
/* if a new cube is being added ...                                                               */
else if (add_cube == TRUE)
	{
	glutSwapBuffers();

	/* determine the current stage in the addition of a new cube and prompt the user for the  */
	/* next vertex in either the right or left image                                          */
	if (left_image_point_a_x == FALSE && left_image_point_a_y == FALSE &&
			left_image_point_b_x == FALSE && left_image_point_b_y == FALSE &&
			left_image_point_c_x == FALSE && left_image_point_c_y == FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x == FALSE && right_image_point_b_x == FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		{
		printf("Digit : add cube ...\n");
		printf("Digit : click on point A in left image ...\n");
		}
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x == FALSE && left_image_point_b_y == FALSE &&
			left_image_point_c_x == FALSE && left_image_point_c_y == FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x == FALSE && right_image_point_b_x == FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point A in right image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x == FALSE && left_image_point_b_y == FALSE &&
			left_image_point_c_x == FALSE && left_image_point_c_y == FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x == FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point B in left image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x != FALSE && left_image_point_b_y != FALSE &&
			left_image_point_c_x == FALSE && left_image_point_c_y == FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x == FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point B in right image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x != FALSE && left_image_point_b_y != FALSE &&
			left_image_point_c_x == FALSE && left_image_point_c_y == FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x != FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point C in left image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x != FALSE && left_image_point_b_y != FALSE &&
			left_image_point_c_x != FALSE && left_image_point_c_y != FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x != FALSE &&
			right_image_point_c_x == FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point C in right image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x != FALSE && left_image_point_b_y != FALSE &&
			left_image_point_c_x != FALSE && left_image_point_c_y != FALSE &&
			left_image_point_d_x == FALSE && left_image_point_d_y == FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x != FALSE &&
			right_image_point_c_x != FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point D in left image ...\n");
	else if (left_image_point_a_x != FALSE && left_image_point_a_y != FALSE &&
			left_image_point_b_x != FALSE && left_image_point_b_y != FALSE &&
			left_image_point_c_x != FALSE && left_image_point_c_y != FALSE &&
			left_image_point_d_x != FALSE && left_image_point_d_y != FALSE &&
			right_image_point_a_x != FALSE && right_image_point_b_x != FALSE &&
			right_image_point_c_x != FALSE && right_image_point_d_x == FALSE)
		printf("Digit : click on point D in right image ...\n");
	}
/* if no new line/edge or cube addition, or image calibration is taking place, redisplay          */
else
	{
	glutSwapBuffers();
	glutPostRedisplay();
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : mouse_func                                                             */
/*         Function Args : button state (up or down) and the screen coordinates when pressed      */
/*         Function Task : this is the registered mouse button callback function in OpenGL and    */
/*                         reacts to a change in state of the mouse buttons                       */
/* ============================================================================================== */

void mouse_func(int button, int down, int x, int y)
{
/* the program only requires the position of a mouse click when calibrating the images, adding    */
/* the vertex of a line/edge, or when adding the vertex of a new cube.  This "if" statement       */
/* controls the collection of mouse coordinates depending on the current state of execution       */

/* calibration of bottom left corner of left image                                                */
if (down == GLUT_UP && left_image_bottom_position == FALSE &&
		left_image_top_position == FALSE && left_image_left_position == FALSE &&
		left_image_right_position == FALSE && right_image_bottom_position == FALSE &&
		right_image_top_position == FALSE && right_image_left_position == FALSE &&
		right_image_right_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	left_image_bottom_position = y;
	left_image_left_position = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* clibration of top right corner of left image                                                   */
else if (down == GLUT_UP && left_image_bottom_position != FALSE &&
		left_image_top_position == FALSE && left_image_left_position != FALSE &&
		left_image_right_position == FALSE && right_image_bottom_position == FALSE &&
		right_image_top_position == FALSE && right_image_left_position == FALSE &&
		right_image_right_position == FALSE)
	{
	printf("        ... mouse position x = %d = y %d\n", x, y);
	left_image_top_position = y;
	left_image_right_position = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* calibration of bottom left corner of right image                                               */
else if (down == GLUT_UP && left_image_bottom_position != FALSE &&
		left_image_top_position != FALSE && left_image_left_position != FALSE &&
		left_image_right_position != FALSE && right_image_bottom_position == FALSE &&
		right_image_top_position == FALSE && right_image_left_position == FALSE &&
		right_image_right_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_bottom_position = y;
	right_image_left_position = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* calibration of top right corner of right image                                                 */
else if (down == GLUT_UP && left_image_bottom_position != FALSE &&
		left_image_top_position != FALSE && left_image_left_position != FALSE &&
		left_image_right_position != FALSE && right_image_bottom_position != FALSE &&
		right_image_top_position == FALSE && right_image_left_position != FALSE &&
		right_image_right_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_top_position = y;
	right_image_right_position = x;
	glutSwapBuffers();
	glutPostRedisplay();
	printf("Digit : ... calibration complete\n");
	calibration = TRUE;                  /* calibration of images is complete so change state */
	}
/* add the start point in the left image of a new line/edge                                       */
else if (down == GLUT_UP && add_line == TRUE && start_left_x_position == FALSE &&
		start_left_y_position == FALSE && start_right_x_position == FALSE &&
		end_left_x_position == FALSE && end_left_y_position == FALSE &&
		end_right_x_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	start_left_x_position = x;
	start_left_y_position = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add the start point in the right image of a new line/edge                                      */
else if (down == GLUT_UP && add_line == TRUE && start_left_x_position != FALSE &&
		start_left_y_position != FALSE && start_right_x_position == FALSE &&
		end_left_x_position == FALSE && end_left_y_position == FALSE &&
		end_right_x_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	start_right_x_position = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add the end point in the left image of a new line/edge                                         */
else if (down == GLUT_UP && add_line == TRUE && start_left_x_position != FALSE &&
		start_left_y_position != FALSE && start_right_x_position != FALSE &&
		end_left_x_position == FALSE && end_left_y_position == FALSE &&
		end_right_x_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	end_left_x_position = x;
	end_left_y_position = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add the end point in the right image of a new line/edge                                        */
else if (down == GLUT_UP && add_line == TRUE && start_left_x_position != FALSE &&
		start_left_y_position != FALSE && start_right_x_position != FALSE &&
		end_left_x_position != FALSE && end_left_y_position != FALSE &&
		end_right_x_position == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	end_right_x_position = x;

	/* all points are collected, ensure they are within the calibration tolerances            */
	if (start_left_x_position >= left_image_left_position &&
		start_left_x_position <= left_image_right_position &&
		start_left_y_position >= left_image_top_position &&
		start_left_y_position <= left_image_bottom_position &&
		start_right_x_position >= right_image_left_position &&
		start_right_x_position <= right_image_right_position &&
		end_left_x_position >= left_image_left_position &&
		end_left_x_position <= left_image_right_position &&
		end_left_y_position >= left_image_top_position &&
		end_left_y_position <= left_image_bottom_position &&
		end_right_x_position >= right_image_left_position &&
		end_right_x_position <= right_image_right_position)

		insert_line(); /* insert the line/edge data into the linked list for display      */
	else
		printf("Error : measurements fell outside image calibration\n");

	start_left_x_position = FALSE; /* reset the temporary values that held the data           */
	start_left_y_position = FALSE;
	start_right_x_position = FALSE;
	end_left_x_position = FALSE;
	end_left_y_position = FALSE;
	end_right_x_position = FALSE;

	add_line = FALSE; /* all data collected so change execution state                         */
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point A in the left image of a new cube                                                    */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x == FALSE &&
		left_image_point_a_y == FALSE && left_image_point_b_x == FALSE &&
		left_image_point_b_y == FALSE && left_image_point_c_x == FALSE &&
		left_image_point_c_y == FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x == FALSE &&
		right_image_point_b_x == FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	left_image_point_a_x = x;
	left_image_point_a_y = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point A in the right image of a new cube                                                   */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x == FALSE &&
		left_image_point_b_y == FALSE && left_image_point_c_x == FALSE &&
		left_image_point_c_y == FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x == FALSE &&
		right_image_point_b_x == FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_point_a_x = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point B in the left image of a new cube                                                    */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x == FALSE &&
		left_image_point_b_y == FALSE && left_image_point_c_x == FALSE &&
		left_image_point_c_y == FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x == FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	left_image_point_b_x = x;
	left_image_point_b_y = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point B in the right image of a new cube                                                   */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x != FALSE &&
		left_image_point_b_y != FALSE && left_image_point_c_x == FALSE &&
		left_image_point_c_y == FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x == FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_point_b_x = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point C in the left image of a new cube                                                    */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x != FALSE &&
		left_image_point_b_y != FALSE && left_image_point_c_x == FALSE &&
		left_image_point_c_y == FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x != FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	left_image_point_c_x = x;
	left_image_point_c_y = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point C in the right image of a new cube                                                   */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x != FALSE &&
		left_image_point_b_y != FALSE && left_image_point_c_x != FALSE &&
		left_image_point_c_y != FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x != FALSE && right_image_point_c_x == FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_point_c_x = x;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point D in the left image of a new cube                                                    */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x != FALSE &&
		left_image_point_b_y != FALSE && left_image_point_c_x != FALSE &&
		left_image_point_c_y != FALSE && left_image_point_d_x == FALSE &&
		left_image_point_d_y == FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x != FALSE && right_image_point_c_x != FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	left_image_point_d_x = x;
	left_image_point_d_y = y;
	glutSwapBuffers();
	glutPostRedisplay();
	}
/* add point D in the right image of a new cube                                                   */
else if (down == GLUT_UP && add_cube == TRUE && left_image_point_a_x != FALSE &&
		left_image_point_a_y != FALSE && left_image_point_b_x != FALSE &&
		left_image_point_b_y != FALSE && left_image_point_c_x != FALSE &&
		left_image_point_c_y != FALSE && left_image_point_d_x != FALSE &&
		left_image_point_d_y != FALSE && right_image_point_a_x != FALSE &&
		right_image_point_b_x != FALSE && right_image_point_c_x != FALSE &&
		right_image_point_d_x == FALSE)
	{
	printf("        ... mouse position x = %d y = %d\n", x, y);
	right_image_point_d_x = x;

	/* all points are collected, ensure they are within the calibration tolerances            */
	if (left_image_point_a_x >= left_image_left_position &&
		left_image_point_a_x <= left_image_right_position &&
		right_image_point_a_x >= right_image_left_position &&
		right_image_point_a_x <= right_image_right_position &&
		left_image_point_b_x >= left_image_left_position &&
		left_image_point_b_x <= left_image_right_position &&
		right_image_point_b_x >= right_image_left_position &&
		right_image_point_b_x <= right_image_right_position &&
		left_image_point_c_x >= left_image_left_position &&
		left_image_point_c_x <= left_image_right_position &&
		right_image_point_c_x >= right_image_left_position &&
		right_image_point_c_x <= right_image_right_position &&
		left_image_point_d_x >= left_image_left_position &&
		left_image_point_d_x <= left_image_right_position &&
		right_image_point_d_x >= right_image_left_position &&
		right_image_point_d_x <= right_image_right_position &&
		left_image_point_a_y >= left_image_top_position &&
		left_image_point_a_y <= left_image_bottom_position &&
		left_image_point_b_y >= left_image_top_position &&
		left_image_point_b_y <= left_image_bottom_position &&
		left_image_point_c_y >= left_image_top_position &&
		left_image_point_c_y <= left_image_bottom_position &&
		left_image_point_d_y >= left_image_top_position &&
		left_image_point_d_y <= left_image_bottom_position)

		insert_square(); /* insert the data into the linked list for displaye             */
	else
		printf("Error : measurements fell outside image calibration\n");

	left_image_point_a_x = FALSE; /* reset the temporary variables that held the data         */
	left_image_point_a_y = FALSE;
	left_image_point_b_x = FALSE;
	left_image_point_b_y = FALSE;
	left_image_point_c_x = FALSE;
	left_image_point_c_y = FALSE;
	left_image_point_d_x = FALSE;
	left_image_point_d_y = FALSE;
	right_image_point_a_x = FALSE;
	right_image_point_b_x = FALSE;
	right_image_point_c_x = FALSE;
	right_image_point_d_x = FALSE;

	add_cube = FALSE; /* all data is collected so change the execution state                  */
	glutSwapBuffers();
	glutPostRedisplay();
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : keyboard_func                                                          */
/*         Function Args : key pressed and the screen coordinates of the mouse when pressed       */
/*         Function Task : this is the registered keyboard callback function in OpenGL and reacts */
/*                         to a key press on the keyboard                                         */
/* ============================================================================================== */

void keyboard_func(unsigned char ch, int x, int y)
{
/* these keys are used to control the execution of the program through user interaction           */

switch(ch)
	{

	case 27 : exit(0); /* terminate the program                                               */
		break;

	case 'a' : if (calibration == TRUE)
			rotate_x = rotate_x + 1 ; /* rotate the camera around x                   */
		break;

	case 'b' : if (calibration == TRUE)
			rotate_y = rotate_y - 1; /* rotate the camera around y                    */
		break;

	case 'c' : if (calibration == TRUE)
			add_cube = TRUE; /* add a cube geometry set                               */
		break;

	case 'd' : depth_factor = depth_factor + 0.1; /* increase the depth factor                */
		   printf("Digit : new depth factor %f\n", depth_factor);
		break;

	case 'f' : depth_factor = depth_factor - 0.1; /* decrease the depth factor                */
		   printf("Digit : new depth factor %f\n", depth_factor);
		break;

	case 'h' : print_help(); /* display the help dialog on the shell                          */
		break;

	case 'l' : if (calibration == TRUE)
			add_line = TRUE; /* add a line/edge geometry set                          */
		break;

	/* print the current camera orientation                                                   */
	case 'o' : printf("Digit : current camera values ...\n");
		   printf("        ... camera x position = %f\n", translate_x);
		   printf("        ... camera y position = %f\n", translate_y);
		   printf("        ... camera z position = %f\n", translate_z);
		   printf("        ... camera x rotation = %f\n", rotate_x);
		   printf("        ... camera y rotation = %f\n", rotate_y);
		   printf("        ... camera zoom       = %f\n", zoom);
		   printf("Digit : ... end of camera values\n");
		break;

	case 'q' : exit(0); /* terminate the program                                              */
		break;

	case 'r' : calibration = FALSE; /* recalibrate the two images                             */
		   left_image_bottom_position = FALSE;
		   left_image_top_position = FALSE;
		   left_image_left_position = FALSE;
		   left_image_right_position = FALSE;
		   right_image_bottom_position = FALSE;
		   right_image_top_position = FALSE;
		   right_image_left_position = FALSE;
		   right_image_right_position = FALSE;
		break;

	case 's' : save_bitmap();
		break;

	case 'v' : if (calibration == TRUE)
			rotate_y = rotate_y + 1; /* rotate the camera around y                    */
		break;

	case 'z' : if (calibration == TRUE)
			rotate_x = rotate_x - 1; /* rotate the camera around x                    */
		break;

	case '=' : if (calibration == TRUE)
			zoom = zoom + 0.01; /* zoom into the scene                                */
		break;

	case '-' : if (calibration == TRUE)
			zoom = zoom - 0.01; /* zoom out of the scene                              */
		break;

	case '.' : scale = scale + 0.1; /* scale the geometry up                                  */
		break;

	case ',' : scale = scale - 0.1; /* scale the geometry down                                */
		break;
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : special_func                                                           */
/*         Function Args : key pressed and the screen coordinates of the mouse when pressed       */
/*         Function Task : this is the registered special callback function in OpenGL and reacts  */
/*                         to a special key press on the keyboard (such as function keys etc)     */
/* ============================================================================================== */

void special_func(int key, int x, int y)
{
/* these special keys are used to pan the virtual camera through the scene                        */

if (calibration == TRUE)
	{
	switch(key)
		{
		case GLUT_KEY_LEFT : translate_x = translate_x + 0.1; /* pan left                 */
			break;

		case GLUT_KEY_RIGHT : translate_x = translate_x - 0.1; /* pan right               */
			break;

		case GLUT_KEY_UP : translate_y = translate_y - 0.1; /* pan up                     */
			break;

		case GLUT_KEY_DOWN : translate_y = translate_y + 0.1; /* pan down                 */
			break;

		case GLUT_KEY_PAGE_UP : translate_z = translate_z + 0.1; /* zoom in               */
			break;

		case GLUT_KEY_PAGE_DOWN : translate_z = translate_z - 0.1; /* zoom out            */
			break;
		}
	}
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : reshape_func                                                           */
/*         Function Args : none                                                                   */
/*         Function Task : this is the registered reshape callback function in OpenGL and reacts  */
/*                         to a change in size of the OpenGL window as a result of user inter-    */
/*                         action                                                                 */
/* ============================================================================================== */

void reshape_func()
{
/* to avoid unnecessary confusion with calls to glRasterPos window resizing has been "disabled"   */
/* for this program, while it is not a specific OpenGL call that disables the resize, when the    */
/* attempts to resize the window, it is simply reset to the default values                        */

glutReshapeWindow(770, 800);
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : initialise_func                                                        */
/*         Function Args : none                                                                   */
/*         Function Task : this function initialises the OpenGL display environment and prepares  */
/*                         the program for writing to the display buffer                          */
/* ============================================================================================== */

void initialise_func(void)
{
glShadeModel(GL_SMOOTH);
glClearColor(0.2f, 0.2f, 0.2f, 1.0f);

axis_display_list(); /* create the display lists                                                  */
grid_display_list();

glEnable(GL_DEPTH_TEST);
glEnable(GL_NORMALIZE);

glMatrixMode(GL_PROJECTION); /* set the projection matrix                                         */
glLoadIdentity();
gluPerspective(50, 1, 0.1, 400);

glMatrixMode(GL_MODELVIEW);
glLoadIdentity();
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         Function Name : main                                                                   */
/*         Function Args : the number of and an array of the command line arguments               */
/*         Function Task : main function of the program than registers all callback functions and */
/*                         places the program in the "main loop"                                  */
/* ============================================================================================== */

int main(int argc, char *argv[])
{
char *left_image_filename;
char *right_image_filename;

int flag_counter;

/* upon program start the command line arguments are parsed to ensure the correct number have     */
/* been received, and program execution continues accordingly                                     */

if (argc == 1)
	{
	/* incorrect command line arguments, terminate execution                                  */
	printf("Usage : digit <left_image_file> <right_image_file>\n");
	printf("      : controls include - \n");
	print_help();
	return 0;
	}
if (argc == 2)
	{
	/* incorrect command line arguments, terminate execution                                  */
	printf("Error : only one image source specified [%s]\n", argv[1]);
	printf("Usage : digit <left_image_file> <right_image_file>\n");
	printf("      : controls include - \n");
	print_help();
	return 0;
	}
if (argc == 3)
	{
	/* correct command line arguments, continue program execution                             */
	printf("Digit : starting, hello\n");
	left_image_filename = argv[1];
	right_image_filename = argv[2];
	printf("Digit : left image source = %s\n", left_image_filename);
	printf("Digit : right image source = %s\n", right_image_filename);
	}
else
	{
	/* incorrect command line arguments, terminate execution                                  */
	printf("Error : invalid flags ");

	for (flag_counter = 3; flag_counter < argc; flag_counter++)
		printf("[%s] ", argv[flag_counter]);

	printf("\n");
	printf("Usage : digit <left_image_file> <right_image_file>\n");
	printf("      : controls include - \n");
	print_help();
	return 0;
	}

atexit(exit_func); /* registration of a function to be called upon execution if exit()            */

/* here we load the bitmap images specified by the user in the command line arguments             */

if (load_images(left_image_filename, right_image_filename) == 0)
	{
	/* images loaded unsuccessfully, so terminate execution                                   */
	printf("Error : unable to load images\n");
	return 0;
	}
else
	/* images loaded OK, continue with execution                                              */
	printf("Digit : images loaded successfully\n");

glutInit(&argc, argv);

/* create the OpenGL window and set-up the display mode                                           */

glutInitDisplayMode(GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE);
glutInitWindowSize(770, 800);
glutInitWindowPosition(255, 0);
glutCreateWindow("Mark Streatfield | Digit");

/* OpenGL callback function registration for the display function, mouse button function,         */
/* keyboard function, special keyboard function and window reshape function                       */

glutDisplayFunc(display_func);
glutMouseFunc(mouse_func);
glutKeyboardFunc(keyboard_func);
glutSpecialFunc(special_func);
glutReshapeFunc(reshape_func);

initialise_func();

glutMainLoop();

return 0;
}

/* ============================================================================================== */
/*         End of Function                                                                        */
/* ============================================================================================== */
/*         End of File                                                                            */
/* ============================================================================================== */
